ssh_hosts = {
    'local_london':'BG-X550',
    'tarik-CCP4stfc':'tarik-CCP4','tarik-CCP4home':'tarik-CCP4',
    'brno':'brno',
    'badb':'badb',
    'badb.rc-harwell.ac.uk':'badb',
    'stfc-cloud':'stfc-cloud',
    'ms01':'ms01',
}

temsim_hosts={
    'BG-X550'   :'/home/ronan/Documents/git/ccp4/src/electron-diffraction/multislice/bin/',
    'tarik-CCP4':'/home/tarik/Documents/git/ccp4/src/electron-diffraction/multislice/bin/',
    'brno'      :'/home/tarik/Documents/git/ccp4/src/electron-diffraction/multislice/bin/',
    'test'      :'/home/lii26466/Documents/git/ccp4/src/electron-diffraction/multislice/bin/',
    'ms01'      :'/home/lii26466/Documents/git/ccp4/src/electron-diffraction/multislice/bin/',
    'badb'      :'/home/lii26466/Documents/git/ccp4/src/electron-diffraction/multislice/bin/',
    'badb.rc-harwell.ac.uk':'/home/lii26466/Documents/git/ccp4/src/electron-diffraction/multislice/bin/',
}
path_hosts={
    'BG-X550'   :'/home/ronan/Documents/git/ccp4/src/electron-diffraction/multislice/dat/',
    'tarik-CCP4':'/home/tarik/Documents/git/ccp4/src/electron-diffraction/multislice/dat/',
    'brno'      :'/home/tarik/Documents/git/ccp4/src/electron-diffraction/multislice/dat/',
    'test'      :'/home/lii26466/Documents/git/ccp4/src/electron-diffraction/multislice/dat/',
    'ms01'      :'/home/lii26466/Documents/git/ccp4/src/electron-diffraction/multislice/dat/',
    'badb'      :'/data3/lii26466/multislice/',
    'badb.rc-harwell.ac.uk' :'/data3/lii26466/multislice/',
}
